#include "audio.h"
#include <list>
#include "main.h"
#include "prog.h"
#include "utils.h"

namespace Audio
{
	static ALCdevice *devhandle;
	static ALCcontext *conhandle;

	ALCdevice *DevHandle()
	{
		return devhandle;
	}
	ALCcontext *ConHandle()
	{
		return conhandle;
	}

	Buffer::Buffer(const char *fname, bool expect_stereo)
	{
		alGenBuffers(1, &buf);

		std::ifstream inp_f(fname, std::ios_base::in | std::ios_base::binary);
		if (!inp_f)
			Main::Exit(Str({"Can't load ", fname, ": File not found."}));
		inp_f >> std::noskipws;

		uint32_t datasize, datasize_alt, freq;
		ALint format;

		uint8_t tmp, *data;

/* 00 */inp_f >> tmp; if (tmp != 'R') Main::Exit(Str({"Can't load ", fname, ": `RIFF` label not found."}));
/* 01 */inp_f >> tmp; if (tmp != 'I') Main::Exit(Str({"Can't load ", fname, ": `RIFF` label not found."}));
/* 02 */inp_f >> tmp; if (tmp != 'F') Main::Exit(Str({"Can't load ", fname, ": `RIFF` label not found."}));
/* 03 */inp_f >> tmp; if (tmp != 'F') Main::Exit(Str({"Can't load ", fname, ": `RIFF` label not found."}));
/* 04 */inp_f >> tmp; datasize = tmp;
/* 05 */inp_f >> tmp; datasize |= tmp << 8;
/* 06 */inp_f >> tmp; datasize |= tmp << 16;
/* 07 */inp_f >> tmp; datasize |= tmp << 24; datasize -= 36;
/* 08 */inp_f >> tmp; if (tmp != 'W') Main::Exit(Str({"Can't load ", fname, ": `WAVE` label not found."}));
/* 09 */inp_f >> tmp; if (tmp != 'A') Main::Exit(Str({"Can't load ", fname, ": `WAVE` label not found."}));
/* 0A */inp_f >> tmp; if (tmp != 'V') Main::Exit(Str({"Can't load ", fname, ": `WAVE` label not found."}));
/* 0B */inp_f >> tmp; if (tmp != 'E') Main::Exit(Str({"Can't load ", fname, ": `WAVE` label not found."}));
/* 0C */inp_f >> tmp; if (tmp != 'f') Main::Exit(Str({"Can't load ", fname, ": `fmt ` label not found."}));
/* 0D */inp_f >> tmp; if (tmp != 'm') Main::Exit(Str({"Can't load ", fname, ": `fmt ` label not found."}));
/* 0E */inp_f >> tmp; if (tmp != 't') Main::Exit(Str({"Can't load ", fname, ": `fmt ` label not found."}));
/* 0F */inp_f >> tmp; if (tmp != ' ') Main::Exit(Str({"Can't load ", fname, ": `fmt ` label not found."}));
/* 10 */inp_f >> tmp; if (tmp != 0x10)Main::Exit(Str({"Can't load ", fname, ": File contains additional data."}));
/* 11 */inp_f >> tmp; if (tmp != 0x00)Main::Exit(Str({"Can't load ", fname, ": File contains additional data."}));
/* 12 */inp_f >> tmp; if (tmp != 0x00)Main::Exit(Str({"Can't load ", fname, ": File contains additional data."}));
/* 13 */inp_f >> tmp; if (tmp != 0x00)Main::Exit(Str({"Can't load ", fname, ": File contains additional data."}));
/* 14 */inp_f >> tmp; if (tmp != 0x01)Main::Exit(Str({"Can't load ", fname, ": File is compressed or uses floating point samples."}));
/* 15 */inp_f >> tmp; if (tmp != 0x00)Main::Exit(Str({"Can't load ", fname, ": File is compressed or uses floating point samples."}));
/* 16 */inp_f >> tmp; if (tmp != expect_stereo + 1)Main::Exit(Str({"Can't load ", fname, expect_stereo ? ": File must contain stereo sound." : ": File must contain mono sound."}));
/* 17 */inp_f >> tmp; if (tmp != 0x00)Main::Exit(Str({"Can't load ", fname, expect_stereo ? ": File must contain stereo sound." : ": File must contain mono sound."}));
/* 18 */inp_f >> tmp; freq = tmp;
/* 19 */inp_f >> tmp; freq |= tmp << 8;
/* 1A */inp_f >> tmp; freq |= tmp << 16;
/* 1B */inp_f >> tmp; freq |= tmp << 24;
/* 1C */inp_f.ignore(6);
/* 1D */
/* 1E */
/* 1F */
/* 20 */
/* 21 */
/* 22 */inp_f >> tmp; switch (tmp) {case 8:  format = expect_stereo ? AL_FORMAT_STEREO8  : AL_FORMAT_MONO8;  break;
									case 16: format = expect_stereo ? AL_FORMAT_STEREO16 : AL_FORMAT_MONO16; break;
									default: Main::Exit(Str({"Can't load ", fname, ": File must contain 8 or 16 bits per sample."})); break;}
/* 23 */inp_f >> tmp; if (tmp != 0x00)Main::Exit(Str({"Can't load ", fname, ": File must contain 8 or 16 bits per sample."}));
/* 24 */inp_f >> tmp; if (tmp != 'd') Main::Exit(Str({"Can't load ", fname, ": `data` label not found."}));
/* 25 */inp_f >> tmp; if (tmp != 'a') Main::Exit(Str({"Can't load ", fname, ": `data` label not found."}));
/* 26 */inp_f >> tmp; if (tmp != 't') Main::Exit(Str({"Can't load ", fname, ": `data` label not found."}));
/* 27 */inp_f >> tmp; if (tmp != 'a') Main::Exit(Str({"Can't load ", fname, ": `data` label not found."}));
/* 28 */inp_f >> tmp; datasize_alt = tmp;
/* 29 */inp_f >> tmp; datasize_alt |= tmp << 8;
/* 2A */inp_f >> tmp; datasize_alt |= tmp << 16;
/* 2B */inp_f >> tmp; datasize_alt |= tmp << 24;

		if (!inp_f)
			Main::Exit(Str({"Can't load ", fname, ": Unexpected end of file (header)."}));

		if (datasize != datasize_alt)
			Main::Exit(Str({"Can't load ", fname, ": File contains controversal information about sound lenght. (", datasize, " and ", datasize_alt, ")"}));

		data = new uint8_t[datasize];

		if (expect_stereo)
		{
			if (datasize % 2)
				Main::Exit(Str({"Can't load ", fname, ": Sound is stereo, but it's size is not multiply of two."}));
			uint32_t max = datasize / 2;
			for (uint32_t i = 0; i < max; i++)
			{
				inp_f >> tmp; ((short *)data)[i] = tmp;
				inp_f >> tmp; ((short *)data)[i] |= tmp << 8;
			}
		}
		else
		{
			for (uint32_t i = 0; i < datasize; i++)
			{
				inp_f >> data[i];
			}
		}
		alBufferData(buf, format, data, datasize, freq);
		delete [] data;

		if (!inp_f)
			Main::Exit(Str({"Can't load ", fname, ": Unexpected end of file."}));
	}



	struct SrcInternal
	{
		ALuint buf, src;
		bool desired_loc; // 0 for unactive, 1 for active
		float volume, pitch;
		bool loop;
		GFX::Vec<float, 3> pos, vel;
		bool rel_location;

		bool change_req;
		bool delete_me;
	};

	static std::list<SrcInternal *> srclist_act, srclist_inact;

	unsigned int ActiveSrcs()
	{
		return srclist_act.size();
	}
	unsigned int InactiveSrcs()
	{
		return srclist_inact.size();
	}

	Source::Source(const Buffer &buf, bool loop, float volume, float pitch, const GFX::Vec<float, 3> &pos, const GFX::Vec<float, 3> &vel, bool rel_location)
	{
		h = new SrcInternal;
		h->buf = buf;
		h->change_req = 0;
		h->desired_loc = 0;
		h->volume = volume;
		h->pitch = pitch;
		h->loop = loop;
		h->pos = pos;
		h->vel = vel;
		h->rel_location = rel_location;
		h->delete_me = 0;

		srclist_inact.push_back(h);
	}

	void Source::Play()                           {h->desired_loc = 1; }
	void Source::Stop()                           {h->desired_loc = 0; }
	void Source::Loop(bool n)                     {h->loop = n;        }
	void Source::Volume(float n)                  {h->volume = n;      }
	void Source::Pitch(float n)                   {h->pitch = n;       }
	void Source::Pos(const GFX::Vec<float, 3> &n) {h->pos = n;         }
	void Source::Vel(const GFX::Vec<float, 3> &n) {h->vel = n;         }
	void Source::RelativeVel(bool n)              {h->rel_location = n;}
	void Source::ApplyChanges()                   {h->change_req = 1;  } // DON'T FORGET TO CALL THIS



	Source::~Source()
	{
		h->delete_me = 1;
	}

	void Tick()
	{
		for (auto it = srclist_act.begin(); it != srclist_act.end();)
		{
			SrcInternal *p = *it;
			if (p->delete_me)
			{
				alDeleteSources(1, &p->src);
				it = srclist_act.erase(it);
				delete p;
				continue;
			}
			if (p->desired_loc == 0)
			{
				alDeleteSources(1, &p->src);
				srclist_inact.push_back(p);
				it = srclist_act.erase(it);
				continue;
			}
			ALint tmp;
			alGetSourcei(p->src, AL_SOURCE_STATE, &tmp);
			if (tmp == AL_INITIAL || tmp == AL_STOPPED)
			{
				p->desired_loc = 0;
				alDeleteSources(1, &p->src);
				srclist_inact.push_back(p);
				it = srclist_act.erase(it);
				continue;
			}
			if (p->change_req == 1)
			{
				p->change_req = 0;
				alSourcef (p->src, AL_GAIN,               p->volume                  );
				alSourcef (p->src, AL_PITCH,              p->pitch                   );
				alSourcei (p->src, AL_LOOPING,            p->loop                    );
				alSourcefv(p->src, AL_POSITION,           (float *) &p->pos          );
				alSourcefv(p->src, AL_VELOCITY,           (float *) &p->vel          );
				alSourcei (p->src, AL_SOURCE_RELATIVE,    p->rel_location            );
				alSourcei (p->src, AL_BUFFER,             p->buf                     );
			}

			it++;
		}

		for (auto it = srclist_inact.begin(); it != srclist_inact.end();)
		{
			SrcInternal *p = *it;
			if (p->delete_me)
			{
				it = srclist_inact.erase(it);
				delete p;
				continue;
			}
			if (srclist_act.size() < MANIFEST_AUDIO_SRCS && p->desired_loc == 1)
			{
				p->change_req = 0;
				alGenSources(1, &p->src);
				alSourcef (p->src, AL_REFERENCE_DISTANCE, MANIFEST_AUDIO_REF_DISTANCE  );
				alSourcef (p->src, AL_MAX_DISTANCE,       MANIFEST_AUDIO_MAX_DISTANCE  );
				alSourcef (p->src, AL_ROLLOFF_FACTOR,     MANIFEST_AUDIO_ROLLOFF_FACTOR);
				alSourcef (p->src, AL_GAIN,               p->volume                    );
				alSourcef (p->src, AL_PITCH,              p->pitch                     );
				alSourcei (p->src, AL_LOOPING,            p->loop                      );
				alSourcefv(p->src, AL_POSITION,           (float *) &p->pos            );
				alSourcefv(p->src, AL_VELOCITY,           (float *) &p->vel            );
				alSourcei (p->src, AL_SOURCE_RELATIVE,    p->rel_location              );
				alSourcei (p->src, AL_BUFFER,             p->buf                       );
				alSourcePlayv(1, &p->src);
				srclist_act.push_back(p);
				it = srclist_inact.erase(it);
				continue;
			}
			it++;
		}
	}

	void Init()
	{
		EXEC_ONCE;

		static constexpr int needed_openal_major = 1, needed_openal_minor = 1; // 1.1

		devhandle = alcOpenDevice(0);
		if (!devhandle)
			Main::Exit("No valid sound device found.");

		ALCint openal_major, openal_minor;
		alcGetIntegerv(devhandle, ALC_MAJOR_VERSION, 1, &openal_major);
		alcGetIntegerv(devhandle, ALC_MINOR_VERSION, 1, &openal_minor);
		if (openal_major < needed_openal_major || (openal_major == needed_openal_major && openal_minor < needed_openal_minor))
			Main::Exit(Str({"OpenAL version runtime check failed. ",needed_openal_major,'.',needed_openal_minor," or newer is needed, but only ",openal_major,'.',openal_minor," is found.\n"
							"Something is wrong with your OpenAL32.dll."}));

		static const ALCint dev_attr_arr[] = {ALC_FREQUENCY, 44100,
		                                      ALC_MONO_SOURCES, MANIFEST_AUDIO_SRCS,
											  ALC_STEREO_SOURCES, MANIFEST_AUDIO_SRCS_STEREO,
											  0                    };
		conhandle = alcCreateContext(devhandle, dev_attr_arr);
		if (!conhandle)
			Main::Exit(Str({"Valid sound device is found, but audio context creation failed. 44100 hz, ", MANIFEST_AUDIO_SRCS, " monosrcs, 0 stereosrcs."}));

		if (!alcMakeContextCurrent(conhandle))
			Main::Exit("Audio context switching failed.");

		alSpeedOfSound(MANIFEST_AUDIO_SPEED_OF_SOUND);
		alDopplerFactor(MANIFEST_AUDIO_DOPPLER_FACTOR);
	}
}