#ifndef AUDIO_H_INCL
#define AUDIO_H_INCL

#include <cstddef>
#include <fstream>
#include <list>
#include "liboal.h"
#include "gfx.h"
#include "main.h"
#include "prog.h"

namespace Audio
{
	ALCdevice *DevHandle();
	ALCcontext *ConHandle();

	inline void ListenerPos(const GFX::Vec<float, 3> &v)
	{
		alListenerfv(AL_POSITION, (float *) &v);
	}

	inline void ListenerVelocity(const GFX::Vec<float, 3> &v)
	{
		alListenerfv(AL_VELOCITY, (float *) &v);
	}

	// X is forward, Z is up
	inline void ListenerRot(const GFX::Quaternion &q)
	{
		q.PrepareMassTransform();
		GFX::Vec<float, 3> arr[2];
		arr[0] = q.MassTransform({1, 0, 0});
		arr[1] = q.MassTransform({0, 0, 1});
		alListenerfv(AL_ORIENTATION, (float *) arr);
	}

	// VECTORS MUST BE ORTHOGONAL, but not necessary normalized. Vectors are IN LOCAL SPACE.
	inline void ListenerRot(const GFX::Vec<float, 3> &forw, const GFX::Vec<float, 3> &up)
	{
		float arr[6] = {forw.x, forw.y, forw.z, up.x, up.y, up.z};
		alListenerfv(AL_ORIENTATION, arr);
	}

	class Buffer
	{
		ALuint buf;
	  public:
		Buffer(const char *fname, bool expect_stereo = 0);
		Buffer(const Buffer &) = delete;
		Buffer(Buffer &&) = delete;
		Buffer &operator=(const Buffer &) = delete;
		Buffer &operator=(Buffer &&) = delete;
		operator ALuint() const {return buf;}

		~Buffer()
		{
			alDeleteBuffers(1, &buf);
		}
	};


	unsigned int ActiveSrcs();
	unsigned int InactiveSrcs();


	struct SrcInternal;

	class Source
	{
		SrcInternal *h;
	  public:
		Source(const Buffer &buf, bool loop, float volume = 1.0, float pitch = 1.0, const GFX::Vec<float, 3> &pos = {0,0,0}, const GFX::Vec<float, 3> &vel = {0,0,0}, bool rel_location = 0);
		Source(const Source &) = delete;
		Source(Source &&) = delete;
		Source &operator=(const Source &) = delete;
		Source &operator=(Source &&) = delete;
		~Source();

		void Play();
		void Stop();

		void Loop(bool n);
		void Volume(float n);
		void Pitch(float n);
		void Pos(const GFX::Vec<float, 3> &n);
		void Vel(const GFX::Vec<float, 3> &n);
		void RelativeVel(bool n);
		void ApplyChanges(); // DON'T FORGET TO CALL THIS
	};

	void Init();
	void Tick();
}

#endif