#include <cmath>

#include "events.h"

#include <ctime>
#include <cstring>
#include <iostream>
#include <map>
#include <string>
#include "libglew.h"
#include "liboal.h"
#include "libsdl.h"
  #define AUDIO_H_TICK_ACCESS
#include "audio.h"
#include "events.h"
  #define GFX_H_RESIZE_ACCESS
#include "gfx.h"
#include "main.h"
#include "utils.h"



namespace Events
{
	//const double target_fps = 9000.0f;
	const double target_fps = 60.0f;
	//const double target_fps = 10.0f;
	//const double target_fps = 5.0f;

	const double target_delta = 1.0 / target_fps;
	static bool relative_mouse = 0;
	static GLenum bufclear_flags = GL_COLOR_BUFFER_BIT;

    int *board, *board_end, boardlen;

    static bool focus = 0;
    static bool winfocus = 0;

    struct MouseData
    {
    	int x, y, rx, ry, mwx, mwy;
    	signed char buttons[32];
    	MouseData() : x(0), y(0), rx(0), ry(0), mwx(0), mwy(0), buttons{} {}
    };

    static MouseData anymouse;
    static std::map<uint32_t, MouseData> mousemap;

    int GetKeyState(unsigned int obj)
    {
    	if (int(obj) >= boardlen)
			return 0;
		else
			return board[obj];
    }

	bool WhatKeyIsPressed(unsigned int *p)
    {
    	for (int *it = board; it != board_end; it++)
		{
			if (*it == 1)
			{
				if (p)
					*p = it - board;
				return 1;
			}
		}
		return 0;
    }

	static double delta, fps;

	double Delta()
	{
		return delta;
	}

	double FPS()
	{
		return fps;
	}


	bool MouseFocus()
	{
		return focus;
	}
	bool WindowFocus()
	{
		return winfocus;
	}

	int MouseX()
	{
		return anymouse.x;
	}
	int MouseY()
	{
		return anymouse.y;
	}
	int MouseRelX()
	{
		return anymouse.rx;
	}
	int MouseRelY()
	{
		return anymouse.ry;
	}
	int MouseWheelX()
	{
		return anymouse.mwx;
	}
	int MouseWheelY()
	{
		return anymouse.mwy;
	}
	int GetMouseButtonState(unsigned int n) // n: 0..31
	{
		return anymouse.buttons[n];
	}

	int MouseX_ID(uint32_t id)
	{
		return mousemap[id].x;
	}
	int MouseY_ID(uint32_t id)
	{
		return mousemap[id].y;
	}
	int MouseRelX_ID(uint32_t id)
	{
		return mousemap[id].rx;
	}
	int MouseRelY_ID(uint32_t id)
	{
		return mousemap[id].ry;
	}
	int MouseWheelX_ID(uint32_t id)
	{
		return mousemap[id].mwx;
	}
	int MouseWheelY_ID(uint32_t id)
	{
		return mousemap[id].mwy;
	}
	int GetMouseButtonState_ID(uint32_t id, unsigned int n) // n: 0..31
	{
		return mousemap[id].buttons[n];
	}

	bool WhatMouseButtonIsPressed(unsigned int *p)
	{
		for (int i = 0; i < 32; i++)
		{
			if (anymouse.buttons[i] == 1)
			{
				if (p)
					*p = i;
				return 1;
			}
		}
		return 0;
	}
	bool WhatMouseButtonIsPressed_ID(uint32_t id, unsigned int *p)
	{
		auto &ref = mousemap[id];
		for (int i = 0; i < 32; i++)
		{
			if (ref.buttons[i] == 1)
			{
				if (p)
					*p = i;
				return 1;
			}
		}
		return 0;
	}
	bool WhatMouseButtonAndIDIsPressed(unsigned int *p, uint32_t *id_p)
	{
		for (auto &it : mousemap)
		{
			for (int i = 0; i < 32; i++)
			{
				if (it.second.buttons[i] == 1)
				{
					if (p)
						*p = i;
					if (id_p)
						*id_p = it.first;
					return 1;
				}
			}
		}
		return 0;
	}
	bool WhatMouseIDHaveButtonPressed(unsigned int n, uint32_t *p)
	{
		for (auto &it : mousemap)
		{
			if (it.second.buttons[n] == 1)
			{
				if (p)
					*p = it.first;
				return 1;
			}
		}
		return 0;
	}

	void ClearDepth(bool n)
	{
		bufclear_flags = n ? GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT : GL_COLOR_BUFFER_BIT;
	}

	void Init()
	{
		SDL_GetKeyboardState(&boardlen);
		board = new int[boardlen]; // Warning: no `delete` for this.
		std::memset(board, 0, boardlen * sizeof (int));
		board_end = board + boardlen;
	}

	namespace MouseMode
	{
		void Show()
		{
			SDL_ShowCursor(1);
		}
		void Hide()
		{
			SDL_ShowCursor(0);
		}

		void Release()
		{
			SDL_SetWindowGrab(GFX::WinHandle(), (SDL_bool)0);
		}
		void Grab()
		{
			SDL_SetWindowGrab(GFX::WinHandle(), (SDL_bool)1);
		}

		void NonRelative() // Also shows cursor (independently from Show()/Hide())
		{
			SDL_SetRelativeMouseMode((SDL_bool)0);
			relative_mouse = 0;
		}
		void Relative() // Also hides cursor (independently from Show()/Hide())
		{
			SDL_SetRelativeMouseMode((SDL_bool)1);
			relative_mouse = 1;
		}
	}

	static inline void FullScreenHotkeyCheck()
	{
		#if MANIFEST_WINDOW_ENABLE_ALT_ENTER == 1
		static bool fullscr = 1;
		#if MANIFEST_WINDOW_RESIZABLE == 1
		int winw = MANIFEST_WINDOW_W, winh = MANIFEST_WINDOW_H;
		#endif
		if (KeyPressed(Key_Enter()) && (KeyDown(Key_Alt_L()) || KeyDown(Key_Alt_R())))
		{
			if (fullscr)
			{
				fullscr = 0;
				SDL_SetWindowFullscreen(GFX::WinHandle(), 0);
				#if MANIFEST_WINDOW_RESIZABLE == 1
				SDL_SetWindowSize(GFX::WinHandle(), winw, winh);
				#endif
			}
			else
			{
				fullscr = 1;
				#if MANIFEST_WINDOW_RESIZABLE == 1
				SDL_GetWindowSize(GFX::WinHandle(), &winw, &winh);
				SDL_DisplayMode mode;
				SDL_GetDesktopDisplayMode(0, &mode);
				SDL_SetWindowSize(GFX::WinHandle(), mode.w, mode.h);
				#endif
				SDL_SetWindowFullscreen(GFX::WinHandle(), SDL_WINDOW_FULLSCREEN);
			}
		}
		#endif
	}

	static inline void EventsTick()
	{
		for (int *it = board; it != board_end; it++)
		{
			switch (*it)
			{
			  case 0:           break;
			  case 1:  *it = 2; break;
			  case 2:           break;
			  default: *it = 0; break;
			}
		}

		anymouse.rx = anymouse.ry = anymouse.mwx = anymouse.mwy = 0;
		for (int i = 0; i < 32; i++)
		{
			switch (anymouse.buttons[i])
			{
			  case 0:                           break;
			  case 1:  anymouse.buttons[i] = 2; break;
			  case 2:                           break;
			  default: anymouse.buttons[i] = 0; break;
			}
		}

		for (auto &it : mousemap)
		{
			it.second.rx = it.second.ry = it.second.mwx = it.second.mwy = 0;
			for (int i = 0; i < 32; i++)
			{
				switch (it.second.buttons[i])
				{
				  case 0:                            break;
				  case 1:  it.second.buttons[i] = 2; break;
				  case 2:                            break;
				  default: it.second.buttons[i] = 0; break;
				}
			}
		}


		SDL_Event curevent;
		while (SDL_PollEvent(&curevent))
		{
			switch (curevent.type)
			{
			  case SDL_QUIT:
				Main::Exit();
				break;
			  case SDL_MOUSEMOTION:
				{
					MouseData &ref = mousemap[curevent.motion.which];
					ref.x = anymouse.x = curevent.motion.x;
					ref.y = anymouse.y = curevent.motion.y;
					if (winfocus)
					{
						anymouse.rx += curevent.motion.xrel;
						ref.rx += curevent.motion.xrel;
						anymouse.ry += curevent.motion.yrel;
						ref.ry += curevent.motion.yrel;
					}
				}
				break;
			  case SDL_MOUSEWHEEL:
				{
					MouseData &ref = mousemap[curevent.wheel.which];
					anymouse.mwx += curevent.wheel.x;
					anymouse.mwy += curevent.wheel.y;
					ref.mwx += curevent.wheel.x;
					ref.mwy += curevent.wheel.y;
				}
				break;
			  case SDL_KEYDOWN:
				if (curevent.key.repeat)
					break;
				board[curevent.key.keysym.scancode] = 1;
				break;
			  case SDL_KEYUP:
				if (curevent.key.repeat)
					break;
				board[curevent.key.keysym.scancode] = 3;
				break;
			  case SDL_MOUSEBUTTONDOWN:
				mousemap[curevent.button.which].buttons[curevent.button.button-1] = anymouse.buttons[curevent.button.button-1] = 1;
				break;
			  case SDL_MOUSEBUTTONUP:
				mousemap[curevent.button.which].buttons[curevent.button.button-1] = anymouse.buttons[curevent.button.button-1] = 3;
				break;
			  case SDL_WINDOWEVENT:
			  	switch (curevent.window.event)
				{
				  case SDL_WINDOWEVENT_FOCUS_LOST:
					if (relative_mouse)
						SDL_SetRelativeMouseMode((SDL_bool)0);
					winfocus = 0;
					break;
				  case SDL_WINDOWEVENT_FOCUS_GAINED:
				  	if (relative_mouse)
						SDL_SetRelativeMouseMode((SDL_bool)1);
					winfocus = 1;
					break;
				}
				break;
			}
		}

		focus = SDL_GetMouseFocus();
	}

	void PreTick()
	{
		GFX::CheckWindowSize();

		glClear(bufclear_flags);

		EventsTick();
		FullScreenHotkeyCheck();
	}

	void PostTick()
	{
		static std::string glerr;
		static bool isglerr;
		while (GLenum err = glGetError())
		{
			isglerr = 1;
			switch (err)
			{
			  case GL_INVALID_ENUM:                  glerr += "GL: Invalid enum.\n";                  break;
			  case GL_INVALID_VALUE:                 glerr += "GL: Invalid value.\n";                 break;
			  case GL_INVALID_OPERATION:             glerr += "GL: Invalid operation.\n";             break;
			  case GL_INVALID_FRAMEBUFFER_OPERATION: glerr += "GL: Invalid framebuffer operation.\n"; break;
			  case GL_OUT_OF_MEMORY:                 glerr += "GL: Out of memory.\n";                 break;
			  case GL_STACK_UNDERFLOW:               glerr += "GL: Stack underflow.\n";               break;
			  case GL_STACK_OVERFLOW:                glerr += "GL: Stack overflow.\n";                break;
			  default:                               glerr += "GL: Unknown error.\n";                 break;
			}
		}
		if (isglerr)
			Main::Exit(glerr.c_str());

		if (ALCenum err = alcGetError(Audio::DevHandle()))
		{
			switch (err)
			{
			  case ALC_INVALID_DEVICE:  Main::Exit("ALC: Invalid device.");
			  case ALC_INVALID_CONTEXT: Main::Exit("ALC: Invalid context.");
			  case ALC_INVALID_ENUM:    Main::Exit("ALC: Invalid enum.");
			  case ALC_INVALID_VALUE:   Main::Exit("ALC: Invalid value.");
			  case ALC_OUT_OF_MEMORY:   Main::Exit("ALC: Out of memory.");
			  default:                  Main::Exit("ALC: Unknown error.");
			}
		}

		SDL_GL_SwapWindow(GFX::WinHandle());

		Audio::Tick();

		static Utils::Timer timer;
		static double prevdelay;
		delta = timer();
		fps = delta > 0.0 ? 1.0 / delta : 9000000;
		if (delta - prevdelay < target_delta)
			Utils::Wait(prevdelay = target_delta - delta + prevdelay);
	}
}