#ifndef EVENTS_H_INCL
#define EVENTS_H_INCL

#include <cstddef>
#include "libglew.h"
#include "prog.h"
#include "main.h"

namespace Events
{
	extern const double target_fps;
	extern const double target_delta;

	/** How to get input IDs:
	 *  Key<'#'>()          # = A..Z || a..z || 0..9 || one_of(`~!@#$%^&*()-=_+,.<>/?;:'"[]{})
	 *  Key_Num<'#'>()      # = 0..9 || one_of(+*-/.,=)
	 *  Key_Num_Enter()
	 *  Key_F<#>()          # = 1..24
	 *  Key_#()             # = one_of(Menu,Esc,Enter,Tab,Backspace,Insert,Delete,Up,Down,Left,Right,PageUp,PageDown,Home,End,CapsLock,ScrollLock,NumLock,PrintScreen,Pause)
	 *  Key_#_$()           # = one_of(Alt,Ctrl,Shift,GUI)      $ = one_of(L,R)
	 *  Mouse_Button<#>()   # = 1..32
	 *  Mouse_Wheel_#()     # = one_of(Up,Down,Left,Right)
	 */
	inline namespace KeyDefs
	{
		#define S(x, y) template <> constexpr inline unsigned int Key<(x)>() {return SDL_SCANCODE_##y;}
		#define SS(x, y, z) S(x, z) S(y, z)
		template <char N> constexpr inline unsigned int Key() // non spec-ed is for `space`
		{
			static_assert(N == ' ', "Input ID is invalid!");
			return SDL_SCANCODE_SPACE;
		}
		SS('0', ')', 0) //
		SS('1', '!', 1) //
		SS('2', '@', 2) //
		SS('3', '#', 3) //
		SS('4', '$', 4) //
		SS('5', '%', 5)  //
		SS('6', '^', 6)  //
		SS('7', '&', 7)  //
		SS('8', '*', 8)  //
		SS('9', '(', 9)  //
		SS('A', 'a', A) //
		SS('B', 'b', B) //
		SS('C', 'c', C) //
		SS('D', 'd', D) //
		SS('E', 'e', E) //
		SS('F', 'f', F)  //
		SS('G', 'g', G)  //
		SS('H', 'h', H)  //
		SS('I', 'i', I)  //
		SS('J', 'j', J)  //
		SS('K', 'k', K) //
		SS('L', 'l', L) //
		SS('M', 'm', M) //
		SS('N', 'n', N) //
		SS('O', 'o', O) //
		SS('P', 'p', P)  //
		SS('Q', 'q', Q)  //
		SS('R', 'r', R)  //
		SS('S', 's', S)  //
		SS('T', 't', T)  //
		SS('U', 'u', U) //
		SS('V', 'v', V) //
		SS('W', 'w', W) //
		SS('X', 'x', X) //
		SS('Y', 'y', Y) //
		SS('Z', 'z', Z)  //
		SS('\'','"', APOSTROPHE)
		SS(',', '<', COMMA)
		SS('.', '>', PERIOD)
		SS('-', '_', MINUS)
		SS('=', '+', EQUALS)
		SS(';', ':', SEMICOLON)
		SS('/', '?', SLASH)
		SS('[', '{', LEFTBRACKET)
		SS(']', '}', RIGHTBRACKET)
		SS('\\','|', BACKSLASH)
		SS('`', '~', GRAVE)
		#undef S
		#undef SS
		template <char N> constexpr inline unsigned int Key_Num() // non spec-ed is for `=`
		{
			static_assert(N == '=', "Input ID is invalid!");
			return SDL_SCANCODE_KP_EQUALS;
		}
		#define S(x, y) template <> constexpr inline unsigned int Key_Num<(x)>() {return SDL_SCANCODE_KP_##y;}
		S('0', 0)
		S('1', 1)
		S('2', 2)
		S('3', 3)
		S('4', 4)
		S('5', 5)
		S('6', 6)
		S('7', 7)
		S('8', 8)
		S('9', 9)
		S('.', DECIMAL)
		S(',', DECIMAL)
		S('+', PLUS)
		S('-', MINUS)
		S('*', MULTIPLY)
		S('/', DIVIDE)
		#undef S
		#define F(x) template <> constexpr inline unsigned int Key_F<(x)>() {return SDL_SCANCODE_F##x;}
		template <unsigned int N> constexpr inline unsigned int Key_F() // non spec-ed is for `F24`
		{
			static_assert(N == 24, "Input ID is invalid!");
			return SDL_SCANCODE_F24;
		}
		F(1 ) F(2 ) F(3 ) F(4 ) F(5 ) F(6 )
		F(7 ) F(8 ) F(9 ) F(10) F(11) F(12)
		F(13) F(14) F(15) F(16) F(17) F(18)
		F(19) F(20) F(21) F(22) F(23)
		#undef F
		#define C(x, y) constexpr inline unsigned int Key_##x() {return SDL_SCANCODE_##y;}
		C(Num_Enter, KP_ENTER)
		C(Menu, MENU)
		C(Esc, ESCAPE)
		C(Enter, RETURN)
		C(Tab, TAB)
		C(Backspace, BACKSPACE)
		C(Insert, INSERT)
		C(Delete, DELETE)
		C(Up, UP)
		C(Down, DOWN)
		C(Left, LEFT)
		C(Right, RIGHT)
		C(PageUp, PAGEUP)
		C(PageDown, PAGEDOWN)
		C(Home, HOME)
		C(End, END)
		C(CapsLock, CAPSLOCK)
		C(ScrollLock, SCROLLLOCK)
		C(NumLock, NUMLOCKCLEAR)
		C(PrintScreen, PRINTSCREEN)
		C(Pause, PAUSE)
		#define CMOD(x, y) C(x##_L, L##y) C(x##_R, R##y)
		CMOD(Alt, ALT)
		CMOD(Ctrl, CTRL)
		CMOD(Shift, SHIFT)
		CMOD(GUI, GUI)
		#undef CMOD
		#undef C
	}

	int GetKeyState(unsigned int n); /// Return value: 0 - up, 1 - pressed, 2 - down, 3 - released.
	bool WhatKeyIsPressed(unsigned int *p); /// Returns 1 if any key is pressed and sets *p to ID of that key, else returns 0 and lefts *p unmodified.

	//                                                                     Time : >--------------->
	//                                                    Key actually pressed? : _ _ # # # # _ _ _
	//                                                            GetKeyState() : 0 0 1 2 2 2 3 0 0
	inline bool KeyUp      (unsigned int n) {return (GetKeyState(n) - 1) & 2;} // # # _ _ _ _ # # #
	inline bool KeyLongUp  (unsigned int n) {return GetKeyState(n) == 0;}      // # # _ _ _ _ _ # #
	inline bool KeyPressed (unsigned int n) {return GetKeyState(n) == 1;}      // _ _ # _ _ _ _ _ _
	inline bool KeyDown    (unsigned int n) {return (GetKeyState(n) + 1) & 2;} // _ _ # # # # _ _ _
	inline bool KeyLongDown(unsigned int n) {return GetKeyState(n) == 2;}      // _ _ _ # # # _ _ _
	inline bool KeyReleased(unsigned int n) {return GetKeyState(n) == 3;}      // _ _ _ _ _ _ # _ _

	bool MouseFocus();
	bool WindowFocus();

	int MouseX();
	int MouseY();
	int MouseRelX();
	int MouseRelY();
	int MouseWheelX();
	int MouseWheelY();
	int GetMouseButtonState(unsigned int n); /// n: 0..31; Return value: 0 - up, 1 - pressed, 2 - down, 3 - released.
	inline bool MouseButtonUp      (unsigned int n) {return (GetMouseButtonState(n) - 1) & 2;}
	inline bool MouseButtonLongUp  (unsigned int n) {return GetMouseButtonState(n) == 0;}
	inline bool MouseButtonPressed (unsigned int n) {return GetMouseButtonState(n) == 1;}
	inline bool MouseButtonDown    (unsigned int n) {return (GetMouseButtonState(n) + 1) & 2;}
	inline bool MouseButtonLongDown(unsigned int n) {return GetMouseButtonState(n) == 2;}
	inline bool MouseButtonReleased(unsigned int n) {return GetMouseButtonState(n) == 3;}

	int MouseX_ID(uint32_t id);
	int MouseY_ID(uint32_t id);
	int MouseRelX_ID(uint32_t id);
	int MouseRelY_ID(uint32_t id);
	int MouseWheelX_ID(uint32_t id);
	int MouseWheelY_ID(uint32_t id);
	int GetMouseButtonState_ID(uint32_t id, unsigned int n); /// n: 0..31; Return value: 0 - up, 1 - pressed, 2 - down, 3 - released.
	inline bool MouseButtonUp_ID      (uint32_t id, unsigned int n) {return (GetMouseButtonState_ID(id, n) - 1) & 2;}
	inline bool MouseButtonLongUp_ID  (uint32_t id, unsigned int n) {return GetMouseButtonState_ID(id, n) == 0;}
	inline bool MouseButtonPressed_ID (uint32_t id, unsigned int n) {return GetMouseButtonState_ID(id, n) == 1;}
	inline bool MouseButtonDown_ID    (uint32_t id, unsigned int n) {return (GetMouseButtonState_ID(id, n) + 1) & 2;}
	inline bool MouseButtonLongDown_ID(uint32_t id, unsigned int n) {return GetMouseButtonState_ID(id, n) == 2;}
	inline bool MouseButtonReleased_ID(uint32_t id, unsigned int n) {return GetMouseButtonState_ID(id, n) == 3;}

	bool WhatMouseButtonIsPressed(unsigned int *p); /// Returns 1 if any button is pressed and sets *p to ID of that button, else returns 0 and lefts *p unmodified.
	bool WhatMouseButtonIsPressed_ID(uint32_t id, unsigned int *p); /// Returns 1 if any button of specific mouse is pressed and sets *p to ID of that button, else returns 0 and lefts *p unmodified.
	bool WhatMouseButtonAndIDIsPressed(unsigned int *p, uint32_t *id_p); /// Returns 1 if any button of any mouse is pressed and sets *p to ID of that button and *id_p to ID of that mouse, else returns 0 and lefts *p unmodified.
	bool WhatMouseIDHaveButtonPressed(unsigned int n, uint32_t *p); /// Returns 1 if specific button is pressed on any mouse and sets *p to ID of that mouse, else returns 0 and lefts *p unmodified.

	double Delta();
	double FPS();
	void PreTick();
	void PostTick();

	void ClearDepth(bool n);

	void Init();

	namespace MouseMode
	{
		void Show();
		void Hide();

		void Release();
		void Grab();

		void NonRelative(); // Also shows cursor (independently from Show()/Hide())
		void Relative();    // Also hides cursor (independently from Show()/Hide())
	}
}

#endif